@@ -1,5 +1,5 @@
 // -*- C++ -*-
-/* Copyright (C) 2005-2019 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -25,8 +25,8 @@
 
 namespace meep {
 
-const int NUM_FIELD_COMPONENTS = 20;
-const int NUM_FIELD_TYPES = 8;
+constexpr int NUM_FIELD_COMPONENTS = 20;
+constexpr int NUM_FIELD_TYPES = 8;
 
 enum component {
   Ex = 0,
@@ -53,7 +53,7 @@ enum component {
   Permeability,
   NO_COMPONENT
 };
-#define Centered Dielectric // better name for centered "dielectric" grid
+const component Centered = Dielectric; // better name for centered "dielectric" grid
 enum derived_component {
   Sx = 100,
   Sy,
@@ -78,16 +78,11 @@ enum field_type {
 enum boundary_side { High = 0, Low };
 enum direction { X = 0, Y, Z, R, P, NO_DIRECTION };
 struct signed_direction {
-  signed_direction(direction dd = X, bool f = false, std::complex<double> ph = 1.0) {
+  explicit signed_direction(direction dd = X, bool f = false, std::complex<double> ph = 1.0) {
     d = dd;
     flipped = f;
     phase = ph;
   };
-  signed_direction(const signed_direction &sd) {
-    d = sd.d;
-    flipped = sd.flipped;
-    phase = sd.phase;
-  }
   signed_direction operator*(std::complex<double> ph);
   bool operator==(const signed_direction &sd) const {
     return (d == sd.d && flipped == sd.flipped && phase == sd.phase);
@@ -153,7 +148,6 @@ component first_field_component(field_type ft);
   for (meep::direction d = dim == meep::Dcyl ? meep::Z : meep::X;                                  \
        d < (dim == meep::Dcyl ? meep::NO_DIRECTION : meep::R); d = meep::direction(d + 1))
 
-// loop over indices idx from is to ie (inclusive) in gv
 #define LOOP_OVER_IVECS(gv, is, ie, idx)                                                           \
   for (ptrdiff_t loop_is1 = (is).yucky_val(0), loop_is2 = (is).yucky_val(1),                       \
                  loop_is3 = (is).yucky_val(2), loop_n1 = ((ie).yucky_val(0) - loop_is1) / 2 + 1,   \
@@ -191,6 +185,68 @@ component first_field_component(field_type ft);
          loop_ibound++)                                                                            \
   LOOP_OVER_IVECS(gv, loop_notowned_is, loop_notowned_ie, idx)
 
+/* The following loop macros work identically to the LOOP_* macros above,
+   but employ shared memory-parallelism using OpenMP. These loops are mainly
+   used in step_generic.cpp and a few other time-critical loops.
+
+   For the parallel implementation, we introduce two dummy loops, one at the beginning
+   and one at the end, in order to "trick" OpenMP to allow us to define our local variables
+   without having to change any other code anywhere else. We can then proceed to do
+   a collapse over all three main loops. */
+
+#define CHUNK_OPENMP _Pragma("omp parallel for")
+
+// the most generic use case where the user
+// can specify a custom clause
+#define PLOOP_OVER_IVECS_C(gv, is, ie, idx, clause)                                                \
+  for (ptrdiff_t loop_is1 = (is).yucky_val(0), loop_is2 = (is).yucky_val(1),                       \
+                 loop_is3 = (is).yucky_val(2), loop_n1 = ((ie).yucky_val(0) - loop_is1) / 2 + 1,   \
+                 loop_n2 = ((ie).yucky_val(1) - loop_is2) / 2 + 1,                                 \
+                 loop_n3 = ((ie).yucky_val(2) - loop_is3) / 2 + 1,                                 \
+                 loop_d1 = (gv).yucky_direction(0), loop_d2 = (gv).yucky_direction(1),             \
+                 loop_d3 = (gv).yucky_direction(2),                                                \
+                 loop_s1 = (gv).stride((meep::direction)loop_d1),                                  \
+                 loop_s2 = (gv).stride((meep::direction)loop_d2),                                  \
+                 loop_s3 = (gv).stride((meep::direction)loop_d3),                                  \
+                 idx0 = (is - (gv).little_corner()).yucky_val(0) / 2 * loop_s1 +                   \
+                        (is - (gv).little_corner()).yucky_val(1) / 2 * loop_s2 +                   \
+                        (is - (gv).little_corner()).yucky_val(2) / 2 * loop_s3,                    \
+                 dummy_first = 0;                                                                  \
+       dummy_first < 1; dummy_first++)                                                             \
+  _Pragma(                                                                                         \
+      clause) for (ptrdiff_t loop_i1 = 0; loop_i1 < loop_n1;                                       \
+                   loop_i1++) for (ptrdiff_t loop_i2 = 0; loop_i2 < loop_n2;                       \
+                                   loop_i2++) for (ptrdiff_t loop_i3 = 0; loop_i3 < loop_n3;       \
+                                                   loop_i3++) for (ptrdiff_t idx =                 \
+                                                                       idx0 + loop_i1 * loop_s1 +  \
+                                                                       loop_i2 * loop_s2 +         \
+                                                                       loop_i3 * loop_s3,          \
+                                                                   dummy_last = 0;                 \
+                                                                   dummy_last < 1; dummy_last++)
+
+// For the main timestepping events, we know
+// we want to do a simple collapse
+#define PLOOP_OVER_IVECS(gv, is, ie, idx)                                                          \
+  PLOOP_OVER_IVECS_C(gv, is, ie, idx, "omp parallel for collapse(3)")
+
+#define PLOOP_OVER_VOL(gv, c, idx)                                                                 \
+  PLOOP_OVER_IVECS(gv, (gv).little_corner() + (gv).iyee_shift(c),                                  \
+                   (gv).big_corner() + (gv).iyee_shift(c), idx)
+
+#define PLOOP_OVER_VOL_OWNED(gv, c, idx)                                                           \
+  PLOOP_OVER_IVECS(gv, (gv).little_owned_corner(c), (gv).big_corner(), idx)
+
+#define PLOOP_OVER_VOL_OWNED0(gv, c, idx)                                                          \
+  PLOOP_OVER_IVECS(gv, (gv).little_owned_corner0(c), (gv).big_corner(), idx)
+
+#define PLOOP_OVER_VOL_NOTOWNED(gv, c, idx)                                                        \
+  for (ivec loop_notowned_is((gv).dim, 0), loop_notowned_ie((gv).dim, 0);                          \
+       loop_notowned_is == zero_ivec((gv).dim);)                                                   \
+    for (int loop_ibound = 0;                                                                      \
+         (gv).get_boundary_icorners(c, loop_ibound, &loop_notowned_is, &loop_notowned_ie);         \
+         loop_ibound++)                                                                            \
+  PLOOP_OVER_IVECS(gv, loop_notowned_is, loop_notowned_ie, idx)
+
 #define LOOPS_ARE_STRIDE1(gv) ((gv).stride((gv).yucky_direction(2)) == 1)
 
 // The following work identically to the LOOP_* macros above,
@@ -204,8 +260,11 @@ component first_field_component(field_type ft);
 // should only use these macros where that is true!  (Basically,
 // all of this is here to support performance hacks of step_generic.)
 
-#if !defined(__INTEL_COMPILER) && !defined(__clang__) && (defined(__GNUC__) || defined(__GNUG__))
+#if !defined(__INTEL_COMPILER) && !defined(__clang__) && !defined(_OPENMP) &&                      \
+    (defined(__GNUC__) || defined(__GNUG__))
 #define IVDEP _Pragma("GCC ivdep")
+#elif defined(_OPENMP)
+#define IVDEP _Pragma("omp simd")
 #elif defined(__INTEL_COMPILER)
 #define IVDEP _Pragma("ivdep")
 #else
@@ -249,31 +308,73 @@ component first_field_component(field_type ft);
          loop_ibound++)                                                                            \
   S1LOOP_OVER_IVECS(gv, loop_notowned_is, loop_notowned_ie, idx)
 
+/* The following is the stride-optimized version of the parallel loops from above.
+   We can use simd vectorization in addition to the usual par for optimization */
+// loop over indices idx from is to ie (inclusive) in gv
+#define PS1LOOP_OVER_IVECS(gv, is, ie, idx)                                                        \
+  for (ptrdiff_t loop_is1 = (is).yucky_val(0), loop_is2 = (is).yucky_val(1),                       \
+                 loop_is3 = (is).yucky_val(2), loop_n1 = ((ie).yucky_val(0) - loop_is1) / 2 + 1,   \
+                 loop_n2 = ((ie).yucky_val(1) - loop_is2) / 2 + 1,                                 \
+                 loop_n3 = ((ie).yucky_val(2) - loop_is3) / 2 + 1,                                 \
+                 loop_d1 = (gv).yucky_direction(0), loop_d2 = (gv).yucky_direction(1),             \
+                 loop_s1 = (gv).stride((meep::direction)loop_d1),                                  \
+                 loop_s2 = (gv).stride((meep::direction)loop_d2), loop_s3 = 1,                     \
+                 idx0 = (is - (gv).little_corner()).yucky_val(0) / 2 * loop_s1 +                   \
+                        (is - (gv).little_corner()).yucky_val(1) / 2 * loop_s2 +                   \
+                        (is - (gv).little_corner()).yucky_val(2) / 2 * loop_s3,                    \
+                 dummy_first = 0;                                                                  \
+       dummy_first < 1; dummy_first++)                                                             \
+  _Pragma("omp parallel for collapse(2)") for (ptrdiff_t loop_i1 = 0; loop_i1 < loop_n1;           \
+                                               loop_i1++) for (ptrdiff_t loop_i2 = 0;              \
+                                                               loop_i2 < loop_n2; loop_i2++)       \
+      _Pragma("omp simd") for (ptrdiff_t loop_i3 = 0; loop_i3 < loop_n3;                           \
+                               loop_i3++) for (ptrdiff_t idx = idx0 + loop_i1 * loop_s1 +          \
+                                                               loop_i2 * loop_s2 + loop_i3,        \
+                                               dummy_last = 0;                                     \
+                                               dummy_last < 1; dummy_last++)
+
+#define PS1LOOP_OVER_VOL(gv, c, idx)                                                               \
+  PS1LOOP_OVER_IVECS(gv, (gv).little_corner() + (gv).iyee_shift(c),                                \
+                     (gv).big_corner() + (gv).iyee_shift(c), idx)
+
+#define PS1LOOP_OVER_VOL_OWNED(gv, c, idx)                                                         \
+  PS1LOOP_OVER_IVECS(gv, (gv).little_owned_corner(c), (gv).big_corner(), idx)
+
+#define PS1LOOP_OVER_VOL_OWNED0(gv, c, idx)                                                        \
+  PS1LOOP_OVER_IVECS(gv, (gv).little_owned_corner0(c), (gv).big_corner(), idx)
+
+#define PS1LOOP_OVER_VOL_NOTOWNED(gv, c, idx)                                                      \
+  for (ivec loop_notowned_is((gv).dim, 0), loop_notowned_ie((gv).dim, 0);                          \
+       loop_notowned_is == meep::zero_ivec((gv).dim);)                                             \
+    for (int loop_ibound = 0;                                                                      \
+         (gv).get_boundary_icorners(c, loop_ibound, &loop_notowned_is, &loop_notowned_ie);         \
+         loop_ibound++)                                                                            \
+  PS1LOOP_OVER_IVECS(gv, loop_notowned_is, loop_notowned_ie, idx)
+
 #define IVEC_LOOP_AT_BOUNDARY                                                                      \
   ((loop_s1 != 0 && (loop_i1 == 0 || loop_i1 == loop_n1 - 1)) ||                                   \
    (loop_s2 != 0 && (loop_i2 == 0 || loop_i2 == loop_n2 - 1)) ||                                   \
    (loop_s3 != 0 && (loop_i3 == 0 || loop_i3 == loop_n3 - 1)))
 
 #define IVEC_LOOP_ILOC(gv, iloc)                                                                   \
-  ivec iloc((gv).dim);                                                                             \
-  iloc.set_direction(direction(loop_d1), loop_is1 + 2 * loop_i1);                                  \
-  iloc.set_direction(direction(loop_d2), loop_is2 + 2 * loop_i2);                                  \
-  iloc.set_direction(direction(loop_d3), loop_is3 + 2 * loop_i3)
+  meep::ivec iloc((gv).dim);                                                                       \
+  iloc.set_direction(meep::direction(loop_d1), loop_is1 + 2 * loop_i1);                            \
+  iloc.set_direction(meep::direction(loop_d2), loop_is2 + 2 * loop_i2);                            \
+  iloc.set_direction(meep::direction(loop_d3), loop_is3 + 2 * loop_i3)
 
 #define IVEC_LOOP_LOC(gv, loc)                                                                     \
-  vec loc((gv).dim);                                                                               \
-  loc.set_direction(direction(loop_d1), (0.5 * loop_is1 + loop_i1) * (gv).inva);                   \
-  loc.set_direction(direction(loop_d2), (0.5 * loop_is2 + loop_i2) * (gv).inva);                   \
-  loc.set_direction(direction(loop_d3), (0.5 * loop_is3 + loop_i3) * (gv).inva)
+  meep::vec loc((gv).dim);                                                                         \
+  loc.set_direction(meep::direction(loop_d1), (0.5 * loop_is1 + loop_i1) * (gv).inva);             \
+  loc.set_direction(meep::direction(loop_d2), (0.5 * loop_is2 + loop_i2) * (gv).inva);             \
+  loc.set_direction(meep::direction(loop_d3), (0.5 * loop_is3 + loop_i3) * (gv).inva)
 
 // integration weight for using LOOP_OVER_IVECS with field::integrate
 #define IVEC_LOOP_WEIGHT1x(s0, s1, e0, e1, i, n, dir)                                              \
   ((i > 1 && i < n - 2)                                                                            \
        ? 1.0                                                                                       \
        : (i == 0 ? (s0).in_direction(meep::direction(dir))                                         \
-                 : (i == 1 ? (s1).in_direction(meep::direction(dir))                               \
-                           : i == n - 1                                                            \
-                                 ? (e0).in_direction(meep::direction(dir))                         \
+                 : (i == 1       ? (s1).in_direction(meep::direction(dir))                         \
+                    : i == n - 1 ? (e0).in_direction(meep::direction(dir))                         \
                                  : (i == n - 2 ? (e1).in_direction(meep::direction(dir)) : 1.0))))
 #define IVEC_LOOP_WEIGHT1(s0, s1, e0, e1, k)                                                       \
   IVEC_LOOP_WEIGHT1x(s0, s1, e0, e1, loop_i##k, loop_n##k, loop_d##k)
@@ -281,6 +382,9 @@ component first_field_component(field_type ft);
   (IVEC_LOOP_WEIGHT1(s0, s1, e0, e1, 3) *                                                          \
    (IVEC_LOOP_WEIGHT1(s0, s1, e0, e1, 2) * ((dV)*IVEC_LOOP_WEIGHT1(s0, s1, e0, e1, 1))))
 
+// equivalent to incrementing a counter, starting from 0, in the loop body
+#define IVEC_LOOP_COUNTER ((loop_i1 * loop_n2 + loop_i2) * loop_n3 + loop_i3)
+
 inline signed_direction flip(signed_direction d) {
   signed_direction d2 = d;
   d2.flipped = !d.flipped;
@@ -305,12 +409,18 @@ inline bool coordinate_mismatch(ndim dim, direction d) {
 
 bool is_tm(component c);
 
+#ifdef SWIG
 extern void abort(const char *, ...); // mympi.cpp
+#else
+[[noreturn]] extern void abort(const char *, ...); // mympi.cpp
+#endif
 
 inline bool is_electric(component c) { return c < Hx; }
 inline bool is_magnetic(component c) { return c >= Hx && c < Dx; }
 inline bool is_D(component c) { return c >= Dx && c < Bx; }
 inline bool is_B(component c) { return c >= Bx && c < Dielectric; }
+inline bool is_E_or_D(component c) { return is_electric(c) || is_D(c); }
+inline bool is_H_or_B(component c) { return is_magnetic(c) || is_B(c); }
 inline bool is_derived(int c) { return c >= Sx; }
 inline bool is_poynting(derived_component c) { return c < EnergyDensity; }
 inline bool is_energydensity(derived_component c) { return c >= EnergyDensity; }
@@ -323,7 +433,7 @@ inline field_type type(component c) {
     return D_stuff;
   else if (is_B(c))
     return B_stuff;
-  abort("Invalid field in type.\n");
+  meep::abort("Invalid field in type.\n");
   return E_stuff; // This is never reached.
 }
 const char *component_name(component c);
@@ -423,14 +533,14 @@ inline component direction_component(component c, direction d) {
   else if (d == NO_DIRECTION && component_direction(c) == d)
     return c;
   else
-    abort("unknown field component %d", c);
+    meep::abort("unknown field component %d", c);
   switch (d) {
     case X: return start_point;
     case Y: return (component)(start_point + 1);
     case Z: return (component)(start_point + 4);
     case R: return (component)(start_point + 2);
     case P: return (component)(start_point + 3);
-    case NO_DIRECTION: abort("vector %d component in NO_DIRECTION", c);
+    case NO_DIRECTION: meep::abort("vector %d component in NO_DIRECTION", c);
   }
   return Ex; // This is never reached.
 }
@@ -441,14 +551,14 @@ inline derived_component direction_component(derived_component c, direction d) {
   else if (is_energydensity(c) && d == NO_DIRECTION)
     return c;
   else
-    abort("unknown field component %d", c);
+    meep::abort("unknown field component %d", c);
   switch (d) {
     case X: return start_point;
     case Y: return (derived_component)(start_point + 1);
     case Z: return (derived_component)(start_point + 4);
     case R: return (derived_component)(start_point + 2);
     case P: return (derived_component)(start_point + 3);
-    case NO_DIRECTION: abort("vector %d derived_component in NO_DIRECTION", c);
+    case NO_DIRECTION: meep::abort("vector %d derived_component in NO_DIRECTION", c);
   }
   return Sx; // This is never reached.
 }
@@ -688,6 +798,12 @@ public:
     return result;
   };
 
+  ivec operator+(int s) const {
+    ivec result = *this;
+    LOOP_OVER_DIRECTIONS(dim, d) result.t[d] += s;
+    return result;
+  };
+
   ivec operator+=(const ivec &a) {
     LOOP_OVER_DIRECTIONS(dim, d) t[d] += a.t[d];
     return *this;
@@ -746,6 +862,19 @@ public:
     return result;
   };
 
+  // element-wise product
+  ivec operator*(const ivec &a) const {
+    ivec result = *this;
+    LOOP_OVER_DIRECTIONS(dim, d) result.t[d] *= a.t[d];
+    return result;
+  };
+
+  ivec operator/(int s) const {
+    ivec result = *this;
+    LOOP_OVER_DIRECTIONS(dim, d) result.t[d] /= s;
+    return result;
+  };
+
   vec operator*(double s) const {
     vec result(dim);
     LOOP_OVER_DIRECTIONS(dim, d) result.set_direction(d, t[d] * s);
@@ -764,10 +893,10 @@ public:
   // case not thread-safe)
   const char *str(char *buffer = 0, size_t buflen = 0);
 
-  ivec round_up_to_even(void) const {
+  ivec round_down_to_even(void) const {
     ivec result(dim);
     LOOP_OVER_DIRECTIONS(dim, d)
-    result.t[d] = t[d] + (t[d] >= 0 ? t[d] : -t[d]) % 2;
+    result.t[d] = t[d] - (t[d] >= 0 ? t[d] : -t[d]) % 2;
     return result;
   }
 
@@ -826,7 +955,6 @@ public:
   };
   volume(const vec &vec1, const vec &vec2);
   volume(const vec &pt);
-  volume(const volume &vol);
   void set_direction_min(direction d, double val) { min_corner.set_direction(d, val); };
   void set_direction_max(direction d, double val) { max_corner.set_direction(d, val); };
   double in_direction_min(direction d) const { return min_corner.in_direction(d); };
@@ -869,6 +997,8 @@ public:
   vec get_max_corner() const { return max_corner; };
   direction normal_direction() const;
 
+  const char *str(char *buffer = 0, size_t buflen = 0);
+
 private:
   vec min_corner, max_corner;
 };
@@ -885,8 +1015,8 @@ class grid_volume {
 public:
   grid_volume(){};
 
-  grid_volume subvolume(ivec is, ivec ie);
-  void init_subvolume(ivec is, ivec ie);
+  grid_volume subvolume(ivec is, ivec ie, component c);
+  void init_subvolume(ivec is, ivec ie, component c);
 
   ndim dim;
   double a, inva /* = 1/a */;
@@ -946,6 +1076,7 @@ public:
   vec loc_at_resolution(ptrdiff_t index, double res) const;
   size_t ntot_at_resolution(double res) const;
   ivec iloc(component, ptrdiff_t index) const;
+  size_t surface_area() const;
 
   ptrdiff_t yee_index(component c) const {
     ptrdiff_t idx = 0;
@@ -973,6 +1104,7 @@ public:
   }
 
   ivec little_owned_corner(component c) const;
+  ivec big_owned_corner(component c) const { return big_corner() - iyee_shift(c); }
   bool owns(const ivec &) const;
   volume surroundings() const;
   volume interior() const;
@@ -986,11 +1118,7 @@ public:
   friend grid_volume vol2d(double xsize, double ysize, double a);
   friend grid_volume vol3d(double xsize, double ysize, double zsize, double a);
 
-  grid_volume split(size_t num, int which) const;
-  grid_volume split_by_effort(int num, int which, int Ngv = 0, const grid_volume *v = NULL,
-                              double *effort = NULL) const;
-  std::vector<grid_volume> split_into_n(int n) const;
-  grid_volume split_at_fraction(bool want_high, int numer, int bestd = -1, int bestlen = 1) const;
+  grid_volume split_at_fraction(bool side_high, int split_pt, int split_dir) const;
   double get_cost() const;
   grid_volume halve(direction d) const;
   void pad_self(direction d);
@@ -1026,10 +1154,14 @@ public:
   double origin_y() const { return origin.y(); }
   double origin_z() const { return origin.z(); }
 
+  const char *str(char *buffer = 0, size_t buflen = 0);
+
+  std::complex<double> get_split_costs(direction d, int split_point, bool frag_cost) const;
+  void tile_split(int &best_split_point, direction &best_split_direction) const;
+  void find_best_split(int desired_chunks, bool frag_cost, int &best_split_point,
+                       direction &best_split_direction, double &left_effort_fraction) const;
+
 private:
-  std::complex<double> get_split_costs(direction d, int split_point) const;
-  void find_best_split(int desired_chunks, int &best_split_point, direction &best_split_direction,
-                       double &left_effort_fraction) const;
   grid_volume(ndim d, double ta, int na, int nb, int nc);
   ivec io;    // integer origin ... always change via set_origin etc.!
   vec origin; // cache of operator[](io), for performance
@@ -1075,23 +1207,24 @@ public:
   int transform(int, int n) const;
   std::complex<double> phase_shift(int, int n) const;
   int multiplicity() const;
+  int multiplicity(ivec &) const;
   bool is_primitive(const ivec &) const;
 
   volume_list *reduce(const volume_list *gl) const;
 
   symmetry operator+(const symmetry &) const;
-  symmetry operator*(std::complex<double>)const;
+  symmetry operator*(std::complex<double>) const;
   symmetry operator-(const symmetry &b) const { return *this + b * (-1.0); }
   symmetry operator-(void) const { return *this * (-1.0); }
   void operator=(const symmetry &);
   bool operator==(const symmetry &) const;
   bool operator!=(const symmetry &S) const { return !(*this == S); };
+  ivec i_symmetry_point;
 
 private:
   signed_direction S[5];
   std::complex<double> ph;
   vec symmetry_point;
-  ivec i_symmetry_point;
   int g; // g is the multiplicity of the symmetry.
   symmetry *next;
   friend symmetry r_to_minus_r_symmetry(double m);
