@@ -2,18 +2,17 @@
 #include "meep_internals.hpp"
 #include "config.h"
 
-#define DPR double *restrict
 #define RPR realnum *restrict
 
-/* These macros get into the guts of the LOOP_OVER_VOL loops to
+/* These macros get into the guts of the PLOOP_OVER_VOL loops to
    efficiently construct the index k into a PML sigma array.
    Basically, k needs to increment by 2 for each increment of one of
    LOOP's for-loops, starting at the appropriate corner of the grid_volume,
    and these macros define the relevant strides etc. for each loop.
    KSTRIDE_DEF defines the relevant strides etc. and goes outside the
    LOOP, wheras KDEF defines the k index and goes inside the LOOP. */
-#define KSTRIDE_DEF(dsig, k, corner)                                                               \
-  const int k##0 = corner.in_direction(dsig) - gv.little_corner().in_direction(dsig);              \
+#define KSTRIDE_DEF(dsig, k, is, gv)                                                               \
+  const int k##0 = is.in_direction(dsig) - gv.little_corner().in_direction(dsig);                  \
   const int s##k##1 = gv.yucky_direction(0) == dsig ? 2 : 0;                                       \
   const int s##k##2 = gv.yucky_direction(1) == dsig ? 2 : 0;                                       \
   const int s##k##3 = gv.yucky_direction(2) == dsig ? 2 : 0
@@ -66,9 +65,11 @@ namespace meep {
 */
 void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
                ptrdiff_t s2, // strides for g1/g2 shift
-               const grid_volume &gv, double dtdx, direction dsig, const DPR sig, const DPR kap,
-               const DPR siginv, RPR fu, direction dsigu, const DPR sigu, const DPR kapu,
-               const DPR siginvu, double dt, const RPR cnd, const RPR cndinv, RPR fcnd) {
+               const grid_volume &gv, const ivec is, const ivec ie, realnum dtdx, direction dsig,
+               const RPR sig, const RPR kap, const RPR siginv, RPR fu, direction dsigu,
+               const RPR sigu, const RPR kapu, const RPR siginvu, realnum dt, const RPR cnd,
+               const RPR cndinv, RPR fcnd) {
+  (void)c;   // currently unused
   if (!g1) { // swap g1 and g2
     SWAP(const RPR, g1, g2);
     SWAP(ptrdiff_t, s1, s2);
@@ -85,38 +86,38 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
   if (dsig == NO_DIRECTION) {    // no PML in f update
     if (dsigu == NO_DIRECTION) { // no fu update
       if (cnd) {
-        double dt2 = dt * 0.5;
+        realnum dt2 = dt * 0.5;
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             f[i] = ((1 - dt2 * cnd[i]) * f[i] - dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2])) *
                    cndinv[i];
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             f[i] = ((1 - dt2 * cnd[i]) * f[i] - dtdx * (g1[i + s1] - g1[i])) * cndinv[i];
           }
         }
       }
       else { // no conductivity
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             f[i] -= dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2]);
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) { f[i] -= dtdx * (g1[i + s1] - g1[i]); }
+          PLOOP_OVER_IVECS(gv, is, ie, i) { f[i] -= dtdx * (g1[i + s1] - g1[i]); }
         }
       }
     }
     else { // fu update, no PML in f update
-      KSTRIDE_DEF(dsigu, ku, gv.little_owned_corner0(c));
+      KSTRIDE_DEF(dsigu, ku, is, gv);
       if (cnd) {
-        double dt2 = dt * 0.5;
+        realnum dt2 = dt * 0.5;
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             fu[i] =
                 ((1 - dt2 * cnd[i]) * fprev - dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2])) *
                 cndinv[i];
@@ -124,9 +125,9 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             fu[i] = ((1 - dt2 * cnd[i]) * fprev - dtdx * (g1[i + s1] - g1[i])) * cndinv[i];
             f[i] = siginvu[ku] * ((kapu[ku] - sigu[ku]) * f[i] + fu[i] - fprev);
           }
@@ -134,17 +135,17 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
       }
       else { // no conductivity
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             fu[i] -= dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2]);
             f[i] = siginvu[ku] * ((kapu[ku] - sigu[ku]) * f[i] + fu[i] - fprev);
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             fu[i] -= dtdx * (g1[i + s1] - g1[i]);
             f[i] = siginvu[ku] * ((kapu[ku] - sigu[ku]) * f[i] + fu[i] - fprev);
           }
@@ -153,12 +154,12 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
     }
   }
   else { /* PML in f update */
-    KSTRIDE_DEF(dsig, k, gv.little_owned_corner0(c));
+    KSTRIDE_DEF(dsig, k, is, gv);
     if (dsigu == NO_DIRECTION) { // no fu update
       if (cnd) {
-        double dt2 = dt * 0.5;
+        realnum dt2 = dt * 0.5;
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             realnum fcnd_prev = fcnd[i];
             fcnd[i] =
@@ -168,7 +169,7 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             realnum fcnd_prev = fcnd[i];
             fcnd[i] = ((1 - dt2 * cnd[i]) * fcnd[i] - dtdx * (g1[i + s1] - g1[i])) * cndinv[i];
@@ -178,14 +179,14 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
       }
       else { // no conductivity (other than PML conductivity)
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             f[i] = ((kap[k] - sig[k]) * f[i] - dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2])) *
                    siginv[k];
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             f[i] = ((kap[k] - sig[k]) * f[i] - dtdx * (g1[i + s1] - g1[i])) * siginv[k];
           }
@@ -193,15 +194,15 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
       }
     }
     else { // fu update + PML in f update
-      KSTRIDE_DEF(dsigu, ku, gv.little_owned_corner0(c));
+      KSTRIDE_DEF(dsigu, ku, is, gv);
       if (cnd) {
-        double dt2 = dt * 0.5;
+        realnum dt2 = dt * 0.5;
         if (g2) {
           //////////////////// MOST GENERAL CASE //////////////////////
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             realnum fcnd_prev = fcnd[i];
             fcnd[i] =
                 ((1 - dt2 * cnd[i]) * fcnd[i] - dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2])) *
@@ -212,10 +213,10 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
           /////////////////////////////////////////////////////////////
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             realnum fcnd_prev = fcnd[i];
             fcnd[i] = ((1 - dt2 * cnd[i]) * fcnd[i] - dtdx * (g1[i + s1] - g1[i])) * cndinv[i];
             fu[i] = ((kap[k] - sig[k]) * fu[i] + (fcnd[i] - fcnd_prev)) * siginv[k];
@@ -225,20 +226,20 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
       }
       else { // no conductivity (other than PML conductivity)
         if (g2) {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             fu[i] = ((kap[k] - sig[k]) * fu[i] - dtdx * (g1[i + s1] - g1[i] + g2[i] - g2[i + s2])) *
                     siginv[k];
             f[i] = siginvu[ku] * ((kapu[ku] - sigu[ku]) * f[i] + fu[i] - fprev);
           }
         }
         else {
-          LOOP_OVER_VOL_OWNED0(gv, c, i) {
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
             DEF_k;
             DEF_ku;
-            double fprev = fu[i];
+            realnum fprev = fu[i];
             fu[i] = ((kap[k] - sig[k]) * fu[i] - dtdx * (g1[i + s1] - g1[i])) * siginv[k];
             f[i] = siginvu[ku] * ((kapu[ku] - sigu[ku]) * f[i] + fu[i] - fprev);
           }
@@ -252,21 +253,22 @@ void step_curl(RPR f, component c, const RPR g1, const RPR g2, ptrdiff_t s1,
    and/or PML).  This is used in 2d calculations to add an exp(i beta z)
    time dependence, which gives an additional i \beta \hat{z} \times
    cross-product in the curl equations. */
-void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, double betadt,
-               direction dsig, const DPR siginv, RPR fu, direction dsigu, const DPR siginvu,
-               const RPR cndinv, RPR fcnd) {
+void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, const ivec is, const ivec ie,
+               realnum betadt, direction dsig, const RPR siginv, RPR fu, direction dsigu,
+               const RPR siginvu, const RPR cndinv, RPR fcnd) {
+  (void)c; // currently unused
   if (!g) return;
   if (dsig != NO_DIRECTION) { // PML in f update
-    KSTRIDE_DEF(dsig, k, gv.little_owned_corner0(c));
+    KSTRIDE_DEF(dsig, k, is, gv);
     if (dsigu != NO_DIRECTION) { // PML in f + fu
-      KSTRIDE_DEF(dsigu, ku, gv.little_owned_corner0(c));
+      KSTRIDE_DEF(dsigu, ku, is, gv);
       if (cndinv) { // conductivity + PML
         //////////////////// MOST GENERAL CASE //////////////////////
-        LOOP_OVER_VOL_OWNED0(gv, c, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_k;
           DEF_ku;
-          double df;
-          double dfcnd = betadt * g[i] * cndinv[i];
+          realnum df;
+          realnum dfcnd = betadt * g[i] * cndinv[i];
           fcnd[i] += dfcnd;
           fu[i] += (df = dfcnd * siginv[k]);
           f[i] += siginvu[ku] * df;
@@ -274,10 +276,10 @@ void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, double be
         /////////////////////////////////////////////////////////////
       }
       else { // PML only
-        LOOP_OVER_VOL_OWNED0(gv, c, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_k;
           DEF_ku;
-          double df;
+          realnum df;
           fu[i] += (df = betadt * g[i] * siginv[k]);
           f[i] += siginvu[ku] * df;
         }
@@ -285,15 +287,15 @@ void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, double be
     }
     else {          // PML in f, no fu
       if (cndinv) { // conductivity + PML
-        LOOP_OVER_VOL_OWNED0(gv, c, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_k;
-          double dfcnd = betadt * g[i] * cndinv[i];
+          realnum dfcnd = betadt * g[i] * cndinv[i];
           fcnd[i] += dfcnd;
           f[i] += dfcnd * siginv[k];
         }
       }
       else { // PML only
-        LOOP_OVER_VOL_OWNED0(gv, c, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_k;
           f[i] += betadt * g[i] * siginv[k];
         }
@@ -302,19 +304,19 @@ void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, double be
   }
   else {                         // no PML in f update
     if (dsigu != NO_DIRECTION) { // fu, no PML in f
-      KSTRIDE_DEF(dsigu, ku, gv.little_owned_corner0(c));
+      KSTRIDE_DEF(dsigu, ku, is, gv);
       if (cndinv) { // conductivity, no PML
-        LOOP_OVER_VOL_OWNED0(gv, c, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_ku;
-          double df;
+          realnum df;
           fu[i] += (df = betadt * g[i] * cndinv[i]);
           f[i] += siginvu[ku] * df;
         }
       }
       else { // no conductivity or PML
-        LOOP_OVER_VOL_OWNED0(gv, c, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_ku;
-          double df;
+          realnum df;
           fu[i] += (df = betadt * g[i]);
           f[i] += siginvu[ku] * df;
         }
@@ -322,10 +324,10 @@ void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, double be
     }
     else {          // no PML, no fu
       if (cndinv) { // conductivity, no PML
-        LOOP_OVER_VOL_OWNED0(gv, c, i) { f[i] += betadt * g[i] * cndinv[i]; }
+        PLOOP_OVER_IVECS(gv, is, ie, i) { f[i] += betadt * g[i] * cndinv[i]; }
       }
       else { // no conductivity or PML
-        LOOP_OVER_VOL_OWNED0(gv, c, i) { f[i] += betadt * g[i]; }
+        PLOOP_OVER_IVECS(gv, is, ie, i) { f[i] += betadt * g[i]; }
       }
     }
   }
@@ -338,10 +340,10 @@ void step_beta(RPR f, component c, const RPR g, const grid_volume &gv, double be
    index change is large, of course, but in that case the chi2/chi3
    power-series expansion isn't accurate anyway, so the cubic isn't
    physical there either. */
-inline double calc_nonlinear_u(const double Dsqr, const double Di, const double chi1inv,
-                               const double chi2, const double chi3) {
-  double c2 = Di * chi2 * (chi1inv * chi1inv);
-  double c3 = Dsqr * chi3 * (chi1inv * chi1inv * chi1inv);
+inline realnum calc_nonlinear_u(const realnum Dsqr, const realnum Di, const realnum chi1inv,
+                                const realnum chi2, const realnum chi3) {
+  realnum c2 = Di * chi2 * (chi1inv * chi1inv);
+  realnum c3 = Dsqr * chi3 * (chi1inv * chi1inv * chi1inv);
   return (1 + c2 + 2 * c3) / (1 + 2 * c2 + 3 * c3);
 }
 
@@ -362,10 +364,11 @@ inline double calc_nonlinear_u(const double Dsqr, const double Di, const double
 
 */
 
-void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const RPR g, const RPR g1,
-                      const RPR g2, const RPR u, const RPR u1, const RPR u2, ptrdiff_t s,
-                      ptrdiff_t s1, ptrdiff_t s2, const RPR chi2, const RPR chi3, RPR fw,
-                      direction dsigw, const DPR sigw, const DPR kapw) {
+void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const ivec is, const ivec ie,
+                      const RPR g, const RPR g1, const RPR g2, const RPR u, const RPR u1,
+                      const RPR u2, ptrdiff_t s, ptrdiff_t s1, ptrdiff_t s2, const RPR chi2,
+                      const RPR chi3, RPR fw, direction dsigw, const RPR sigw, const RPR kapw) {
+  (void)fc; // currently unused
   if (!f) return;
 
   if ((!g1 && g2) || (g1 && g2 && !u1 && u2)) { /* swap g1 and g2 */
@@ -374,7 +377,7 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const RPR g, c
     SWAP(ptrdiff_t, s1, s2);
   }
 
-  // stable averaging of offdiagonal components
+// stable averaging of offdiagonal components
 #define OFFDIAG(u, g, sx)                                                                          \
   (0.25 * ((g[i] + g[i - sx]) * u[i] + (g[i + s] + g[(i + s) - sx]) * u[i + s]))
 
@@ -382,17 +385,17 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const RPR g, c
      of the "MOST GENERAL CASE" loop with various terms thrown out. */
 
   if (dsigw != NO_DIRECTION) { //////// PML case (with fw) /////////////
-    KSTRIDE_DEF(dsigw, kw, gv.little_owned_corner0(fc));
+    KSTRIDE_DEF(dsigw, kw, is, gv);
     if (u1 && u2) { // 3x3 off-diagonal u
       if (chi3) {
         //////////////////// MOST GENERAL CASE //////////////////////
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-          double g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+          realnum g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
+          realnum gs = g[i];
+          realnum us = u[i];
           DEF_kw;
-          double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+          realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
           fw[i] = (gs * us + OFFDIAG(u1, g1, s1) + OFFDIAG(u2, g2, s2)) *
                   calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s + g2s * g2s), gs, us, chi2[i],
                                    chi3[i]);
@@ -401,11 +404,11 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const RPR g, c
         /////////////////////////////////////////////////////////////
       }
       else {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum gs = g[i];
+          realnum us = u[i];
           DEF_kw;
-          double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+          realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
           fw[i] = (gs * us + OFFDIAG(u1, g1, s1) + OFFDIAG(u2, g2, s2));
           f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
         }
@@ -413,86 +416,84 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const RPR g, c
     }
     else if (u1) { // 2x2 off-diagonal u
       if (chi3) {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+          realnum gs = g[i];
+          realnum us = u[i];
           DEF_kw;
-          double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+          realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
           fw[i] = (gs * us + OFFDIAG(u1, g1, s1)) *
                   calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s), gs, us, chi2[i], chi3[i]);
           f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
         }
       }
       else {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum gs = g[i];
+          realnum us = u[i];
           DEF_kw;
-          double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+          realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
           fw[i] = (gs * us + OFFDIAG(u1, g1, s1));
           f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
         }
       }
     }
     else if (u2) { // 2x2 off-diagonal u
-      abort("bug - didn't swap off-diagonal terms!?");
+      meep::abort("bug - didn't swap off-diagonal terms!?");
     }
     else { // diagonal u
       if (chi3) {
         if (g1 && g2) {
-          LOOP_OVER_VOL_OWNED(gv, fc, i) {
-            double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-            double g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
-            double gs = g[i];
-            double us = u[i];
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+            realnum g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
+            realnum gs = g[i];
+            realnum us = u[i];
             DEF_kw;
-            double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+            realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
             fw[i] = (gs * us) * calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s + g2s * g2s), gs, us,
                                                  chi2[i], chi3[i]);
             f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
           }
         }
         else if (g1) {
-          LOOP_OVER_VOL_OWNED(gv, fc, i) {
-            double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-            double gs = g[i];
-            double us = u[i];
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+            realnum gs = g[i];
+            realnum us = u[i];
             DEF_kw;
-            double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+            realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
             fw[i] = (gs * us) *
                     calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s), gs, us, chi2[i], chi3[i]);
             f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
           }
         }
-        else if (g2) {
-          abort("bug - didn't swap off-diagonal terms!?");
-        }
+        else if (g2) { meep::abort("bug - didn't swap off-diagonal terms!?"); }
         else {
-          LOOP_OVER_VOL_OWNED(gv, fc, i) {
-            double gs = g[i];
-            double us = u[i];
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum gs = g[i];
+            realnum us = u[i];
             DEF_kw;
-            double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+            realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
             fw[i] = (gs * us) * calc_nonlinear_u(gs * gs, gs, us, chi2[i], chi3[i]);
             f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
           }
         }
       }
       else if (u) {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum gs = g[i];
+          realnum us = u[i];
           DEF_kw;
-          double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+          realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
           fw[i] = (gs * us);
           f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
         }
       }
       else {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
           DEF_kw;
-          double fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
+          realnum fwprev = fw[i], kapwkw = kapw[kw], sigwkw = sigw[kw];
           fw[i] = g[i];
           f[i] += (kapwkw + sigwkw) * fw[i] - (kapwkw - sigwkw) * fwprev;
         }
@@ -502,86 +503,84 @@ void step_update_EDHB(RPR f, component fc, const grid_volume &gv, const RPR g, c
   else {            /////////////// no PML (no fw) ///////////////////
     if (u1 && u2) { // 3x3 off-diagonal u
       if (chi3) {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-          double g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+          realnum g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
+          realnum gs = g[i];
+          realnum us = u[i];
           f[i] = (gs * us + OFFDIAG(u1, g1, s1) + OFFDIAG(u2, g2, s2)) *
                  calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s + g2s * g2s), gs, us, chi2[i],
                                   chi3[i]);
         }
       }
       else {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum gs = g[i];
+          realnum us = u[i];
           f[i] = (gs * us + OFFDIAG(u1, g1, s1) + OFFDIAG(u2, g2, s2));
         }
       }
     }
     else if (u1) { // 2x2 off-diagonal u
       if (chi3) {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+          realnum gs = g[i];
+          realnum us = u[i];
           f[i] = (gs * us + OFFDIAG(u1, g1, s1)) *
                  calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s), gs, us, chi2[i], chi3[i]);
         }
       }
       else {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum gs = g[i];
+          realnum us = u[i];
           f[i] = (gs * us + OFFDIAG(u1, g1, s1));
         }
       }
     }
     else if (u2) { // 2x2 off-diagonal u
-      abort("bug - didn't swap off-diagonal terms!?");
+      meep::abort("bug - didn't swap off-diagonal terms!?");
     }
     else { // diagonal u
       if (chi3) {
         if (g1 && g2) {
-          LOOP_OVER_VOL_OWNED(gv, fc, i) {
-            double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-            double g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
-            double gs = g[i];
-            double us = u[i];
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+            realnum g2s = g2[i] + g2[i + s] + g2[i - s2] + g2[i + (s - s2)];
+            realnum gs = g[i];
+            realnum us = u[i];
             f[i] = (gs * us) * calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s + g2s * g2s), gs, us,
                                                 chi2[i], chi3[i]);
           }
         }
         else if (g1) {
-          LOOP_OVER_VOL_OWNED(gv, fc, i) {
-            double g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
-            double gs = g[i];
-            double us = u[i];
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum g1s = g1[i] + g1[i + s] + g1[i - s1] + g1[i + (s - s1)];
+            realnum gs = g[i];
+            realnum us = u[i];
             f[i] = (gs * us) *
                    calc_nonlinear_u(gs * gs + 0.0625 * (g1s * g1s), gs, us, chi2[i], chi3[i]);
           }
         }
-        else if (g2) {
-          abort("bug - didn't swap off-diagonal terms!?");
-        }
+        else if (g2) { meep::abort("bug - didn't swap off-diagonal terms!?"); }
         else {
-          LOOP_OVER_VOL_OWNED(gv, fc, i) {
-            double gs = g[i];
-            double us = u[i];
+          PLOOP_OVER_IVECS(gv, is, ie, i) {
+            realnum gs = g[i];
+            realnum us = u[i];
             f[i] = (gs * us) * calc_nonlinear_u(gs * gs, gs, us, chi2[i], chi3[i]);
           }
         }
       }
       else if (u) {
-        LOOP_OVER_VOL_OWNED(gv, fc, i) {
-          double gs = g[i];
-          double us = u[i];
+        PLOOP_OVER_IVECS(gv, is, ie, i) {
+          realnum gs = g[i];
+          realnum us = u[i];
           f[i] = (gs * us);
         }
       }
       else
-        LOOP_OVER_VOL_OWNED(gv, fc, i) { f[i] = g[i]; }
+        PLOOP_OVER_IVECS(gv, is, ie, i) { f[i] = g[i]; }
     }
   }
 }
