@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -124,14 +124,14 @@ component first_field_component(field_type ft) {
     case H_stuff: return Hx;
     case D_stuff: return Dx;
     case B_stuff: return Bx;
-    default: abort("bug - only E/H/D/B stuff have components"); return NO_COMPONENT;
+    default: meep::abort("bug - only E/H/D/B stuff have components"); return NO_COMPONENT;
   }
 }
 
 vec min(const vec &vec1, const vec &vec2) {
   vec m(vec1.dim);
   LOOP_OVER_DIRECTIONS(vec1.dim, d) {
-    m.set_direction(d, min(vec1.in_direction(d), vec2.in_direction(d)));
+    m.set_direction(d, std::min(vec1.in_direction(d), vec2.in_direction(d)));
   }
   return m;
 }
@@ -139,7 +139,7 @@ vec min(const vec &vec1, const vec &vec2) {
 vec max(const vec &vec1, const vec &vec2) {
   vec m(vec1.dim);
   LOOP_OVER_DIRECTIONS(vec1.dim, d) {
-    m.set_direction(d, max(vec1.in_direction(d), vec2.in_direction(d)));
+    m.set_direction(d, std::max(vec1.in_direction(d), vec2.in_direction(d)));
   }
   return m;
 }
@@ -147,7 +147,7 @@ vec max(const vec &vec1, const vec &vec2) {
 ivec min(const ivec &ivec1, const ivec &ivec2) {
   ivec m(ivec1.dim);
   LOOP_OVER_DIRECTIONS(ivec1.dim, d) {
-    m.set_direction(d, min(ivec1.in_direction(d), ivec2.in_direction(d)));
+    m.set_direction(d, std::min(ivec1.in_direction(d), ivec2.in_direction(d)));
   }
   return m;
 }
@@ -155,7 +155,7 @@ ivec min(const ivec &ivec1, const ivec &ivec2) {
 ivec max(const ivec &ivec1, const ivec &ivec2) {
   ivec m(ivec1.dim);
   LOOP_OVER_DIRECTIONS(ivec1.dim, d) {
-    m.set_direction(d, max(ivec1.in_direction(d), ivec2.in_direction(d)));
+    m.set_direction(d, std::max(ivec1.in_direction(d), ivec2.in_direction(d)));
   }
   return m;
 }
@@ -172,9 +172,6 @@ volume::volume(const vec &pt) {
   max_corner = pt;
 }
 
-volume::volume(const volume &vol)
-    : dim(vol.dim), min_corner(vol.min_corner), max_corner(vol.max_corner) {}
-
 double volume::computational_volume() const {
   double vol = 1.0;
   LOOP_OVER_DIRECTIONS(dim, d) { vol *= in_direction(d); }
@@ -198,16 +195,16 @@ double volume::full_volume() const {
 
 double volume::diameter() const {
   double diam = 0.0;
-  LOOP_OVER_DIRECTIONS(dim, d) { diam = max(diam, in_direction(d)); }
+  LOOP_OVER_DIRECTIONS(dim, d) { diam = std::max(diam, in_direction(d)); }
   return diam;
 }
 
 volume volume::intersect_with(const volume &a) const {
-  if (a.dim != dim) abort("Can't intersect volumes of dissimilar dimensions.\n");
+  if (a.dim != dim) meep::abort("Can't intersect volumes of dissimilar dimensions.\n");
   volume result(dim);
   LOOP_OVER_DIRECTIONS(dim, d) {
-    double minval = max(in_direction_min(d), a.in_direction_min(d));
-    double maxval = min(in_direction_max(d), a.in_direction_max(d));
+    double minval = std::max(in_direction_min(d), a.in_direction_min(d));
+    double maxval = std::min(in_direction_max(d), a.in_direction_max(d));
     if (minval > maxval) return volume(zero_vec(dim), zero_vec(dim));
     result.set_direction_min(d, minval);
     result.set_direction_max(d, maxval);
@@ -216,10 +213,10 @@ volume volume::intersect_with(const volume &a) const {
 }
 
 bool volume::intersects(const volume &a) const {
-  if (a.dim != dim) abort("Can't intersect volumes of dissimilar dimensions.\n");
+  if (a.dim != dim) meep::abort("Can't intersect volumes of dissimilar dimensions.\n");
   LOOP_OVER_DIRECTIONS(dim, d) {
-    double minval = max(in_direction_min(d), a.in_direction_min(d));
-    double maxval = min(in_direction_max(d), a.in_direction_max(d));
+    double minval = std::max(in_direction_min(d), a.in_direction_min(d));
+    double maxval = std::min(in_direction_max(d), a.in_direction_max(d));
     if (minval > maxval) return false;
   }
   return true;
@@ -321,7 +318,7 @@ component grid_volume::eps_component() const {
     case D3: return Dielectric;
     case Dcyl: return Hp;
   }
-  abort("Unsupported dimensionality eps.\n");
+  meep::abort("Unsupported dimensionality eps.\n");
   return Ex;
 }
 
@@ -337,7 +334,7 @@ void grid_volume::yee2cent_offsets(component c, ptrdiff_t &offset1, ptrdiff_t &o
   offset1 = offset2 = 0;
   LOOP_OVER_DIRECTIONS(dim, d) {
     if (!iyee_shift(c).in_direction(d)) {
-      if (offset2) abort("weird yee shift for component %s", component_name(c));
+      if (offset2) meep::abort("weird yee shift for component %s", component_name(c));
       if (offset1)
         offset2 = stride(d);
       else
@@ -457,14 +454,10 @@ bool grid_volume::owns(const ivec &p) const {
     return o.x() > 0 && o.x() <= nx() * 2 && o.y() > 0 && o.y() <= ny() * 2 && o.z() > 0 &&
            o.z() <= nz() * 2;
   }
-  else if (dim == D2) {
-    return o.x() > 0 && o.x() <= nx() * 2 && o.y() > 0 && o.y() <= ny() * 2;
-  }
-  else if (dim == D1) {
-    return o.z() > 0 && o.z() <= nz() * 2;
-  }
+  else if (dim == D2) { return o.x() > 0 && o.x() <= nx() * 2 && o.y() > 0 && o.y() <= ny() * 2; }
+  else if (dim == D1) { return o.z() > 0 && o.z() <= nz() * 2; }
   else {
-    abort("Unsupported dimension in owns.\n");
+    meep::abort("Unsupported dimension in owns.\n");
     return false;
   }
 }
@@ -547,7 +540,7 @@ void grid_volume::interpolate(component c, const vec &p, ptrdiff_t indices[8],
     printf("Or in other words... %g %g\n", operator[](locs[0]).r(), operator[](locs[0]).z());
     printf("I %s own the interpolated point.\n", owns(locs[0]) ? "actually" : "don't");
     print();
-    abort("Error made in interpolation of %s--fix this bug!!!\n", component_name(c));
+    meep::abort("Error made in interpolation of %s--fix this bug!!!\n", component_name(c));
   }
   // Throw out out of range indices:
   for (int i = 0; i < 8 && weights[i]; i++)
@@ -558,7 +551,7 @@ void grid_volume::interpolate(component c, const vec &p, ptrdiff_t indices[8],
     printf("Error at point %g %g\n", p.r(), p.z());
     printf("Interpolated to point %d %d\n", locs[0].r(), locs[0].z());
     print();
-    abort("Error made in interpolation of %s--fix this bug!!!\n", component_name(c));
+    meep::abort("Error made in interpolation of %s--fix this bug!!!\n", component_name(c));
   }
 }
 
@@ -596,7 +589,7 @@ void grid_volume::interpolate(component c, const vec &pc, ivec locs[8], double w
   for (int i = 0; i < already_have; i++) {
     if (weights[i] < 0.0) {
       if (-weights[i] >= SMALL * 1e5)
-        abort("large negative interpolation weight[%d] = %e\n", i, weights[i]);
+        meep::abort("large negative interpolation weight[%d] = %e\n", i, weights[i]);
       weights[i] = 0.0;
     }
     else if (weights[i] < SMALL)
@@ -678,7 +671,7 @@ double grid_volume::zmin() const {
 double grid_volume::rmax() const {
   const double qinva = 0.25 * inva;
   if (dim == Dcyl) return origin.r() + nr() * inva + qinva;
-  abort("No rmax in these dimensions.\n");
+  meep::abort("No rmax in these dimensions.\n");
   return 0.0; // This is never reached.
 }
 
@@ -686,11 +679,9 @@ double grid_volume::rmin() const {
   const double qinva = 0.25 * inva;
   if (dim == Dcyl) {
     if (origin.r() == 0.0) { return 0.0; }
-    else {
-      return origin.r() + qinva;
-    }
+    else { return origin.r() + qinva; }
   }
-  abort("No rmin in these dimensions.\n");
+  meep::abort("No rmin in these dimensions.\n");
   return 0.0; // This is never reached.
 }
 
@@ -709,8 +700,8 @@ double grid_volume::boundary_location(boundary_side b, direction d) const {
           return loc(Ep, ntot() - 1).z();
         else
           return loc(Ex, ntot() - 1).z();
-      case P: abort("P has no boundary!\n");
-      case NO_DIRECTION: abort("NO_DIRECTION has no boundary!\n");
+      case P: meep::abort("P has no boundary!\n");
+      case NO_DIRECTION: meep::abort("NO_DIRECTION has no boundary!\n");
     }
   else
     switch (d) {
@@ -722,8 +713,8 @@ double grid_volume::boundary_location(boundary_side b, direction d) const {
           return loc(Ep, 0).z();
         else
           return loc(Ex, 0).z();
-      case P: abort("P has no boundary!\n");
-      case NO_DIRECTION: abort("NO_DIRECTION has no boundary!\n");
+      case P: meep::abort("P has no boundary!\n");
+      case NO_DIRECTION: meep::abort("NO_DIRECTION has no boundary!\n");
     }
   return 0.0;
 }
@@ -760,8 +751,8 @@ bool grid_volume::intersect_with(const grid_volume &vol_in, grid_volume *interse
   int temp_num[3] = {0, 0, 0};
   ivec new_io(dim);
   LOOP_OVER_DIRECTIONS(dim, d) {
-    int minval = max(little_corner().in_direction(d), vol_in.little_corner().in_direction(d));
-    int maxval = min(big_corner().in_direction(d), vol_in.big_corner().in_direction(d));
+    int minval = std::max(little_corner().in_direction(d), vol_in.little_corner().in_direction(d));
+    int maxval = std::min(big_corner().in_direction(d), vol_in.big_corner().in_direction(d));
     if (minval >= maxval) return false;
     temp_num[d % 3] = (maxval - minval) / 2;
     new_io.set_direction(d, minval);
@@ -787,7 +778,7 @@ bool grid_volume::intersect_with(const grid_volume &vol_in, grid_volume *interse
         vol_containing.shift_origin(d, thick * 2);
         vol_containing.set_num_direction(d, vol_containing.num_direction(d) - thick);
         if (vol_containing.little_corner().in_direction(d) < vol_in.little_corner().in_direction(d))
-          abort("intersect_with: little corners differ by odd integer?");
+          meep::abort("intersect_with: little corners differ by odd integer?");
       }
       if (vol_containing.big_corner().in_direction(d) > vol_in.big_corner().in_direction(d)) {
         // shave off upper slice from vol_containing and add it to others
@@ -800,7 +791,7 @@ bool grid_volume::intersect_with(const grid_volume &vol_in, grid_volume *interse
         counter++;
         vol_containing.set_num_direction(d, vol_containing.num_direction(d) - thick);
         if (vol_containing.big_corner().in_direction(d) < vol_in.big_corner().in_direction(d))
-          abort("intersect_with: big corners differ by odd integer?");
+          meep::abort("intersect_with: big corners differ by odd integer?");
       }
     }
     *num_others = counter;
@@ -816,8 +807,8 @@ bool grid_volume::intersect_with(const grid_volume &vol_in, grid_volume *interse
       final_points += temp;
     }
     if (initial_points != final_points)
-      abort("intersect_with: initial_points != final_points,  %zd, %zd\n", initial_points,
-            final_points);
+      meep::abort("intersect_with: initial_points != final_points,  %zd, %zd\n", initial_points,
+                  final_points);
   }
   return true;
 }
@@ -828,7 +819,7 @@ vec grid_volume::loc_at_resolution(ptrdiff_t index, double res) const {
     const direction d = (direction)dd;
     if (has_boundary(High, d)) {
       const double dist = boundary_location(High, d) - boundary_location(Low, d);
-      const int nhere = max(1, (int)floor(dist * res + 0.5));
+      const int nhere = std::max(1, (int)floor(dist * res + 0.5));
       where.set_direction(d, origin.in_direction(d) + ((index % nhere) + 0.5) * (1.0 / res));
       index /= nhere;
     }
@@ -842,7 +833,7 @@ size_t grid_volume::ntot_at_resolution(double res) const {
     if (has_boundary(High, (direction)d)) {
       const double dist =
           boundary_location(High, (direction)d) - boundary_location(Low, (direction)d);
-      mytot *= max(size_t(1), (size_t)(dist * res + 0.5));
+      mytot *= std::max(size_t(1), (size_t)(dist * res + 0.5));
     }
   return mytot;
 }
@@ -861,11 +852,11 @@ ivec grid_volume::iloc(component c, ptrdiff_t ind) const {
 }
 
 size_t grid_volume::surface_area() const {
-  switch(dim) {
-    case Dcyl: return 2*(nr()+nz());
-    case D1: 2;
-    case D2: return 2*(nx()+ny());
-    case D3: return 2*(nx()*ny()+nx()*nz()+ny()*nz());
+  switch (dim) {
+    case Dcyl: return 2 * (nr() + nz());
+    case D1: return 2;
+    case D2: return 2 * (nx() + ny());
+    case D3: return 2 * (nx() * ny() + nx() * nz() + ny() * nz());
   }
   return 0; // This is never reached.
 }
@@ -875,7 +866,7 @@ vec grid_volume::dr() const {
     case Dcyl: return veccyl(inva, 0.0);
     case D1:
     case D2:
-    case D3: abort("Error in dr\n");
+    case D3: meep::abort("Error in dr\n");
   }
   return vec(0); // This is never reached.
 }
@@ -885,7 +876,7 @@ vec grid_volume::dx() const {
     case D3: return vec(inva, 0, 0);
     case D2: return vec(inva, 0);
     case D1:
-    case Dcyl: abort("Error in dx.\n");
+    case Dcyl: meep::abort("Error in dx.\n");
   }
   return vec(0); // This is never reached.
 }
@@ -895,7 +886,7 @@ vec grid_volume::dy() const {
     case D3: return vec(0, inva, 0);
     case D2: return vec(0, inva);
     case D1:
-    case Dcyl: abort("Error in dy.\n");
+    case Dcyl: meep::abort("Error in dy.\n");
   }
   return vec(0); // This is never reached.
 }
@@ -905,22 +896,22 @@ vec grid_volume::dz() const {
     case Dcyl: return veccyl(0.0, inva);
     case D3: return vec(0, 0, inva);
     case D1: return vec(inva);
-    case D2: abort("dz doesn't exist in 2D\n");
+    case D2: meep::abort("dz doesn't exist in 2D\n");
   }
   return vec(0); // This is never reached.
 }
 
 grid_volume volone(double zsize, double a) {
   if (!isinteger(zsize * a))
-    master_printf_stderr(
-      "Warning: grid volume is not an integer number of pixels; cell size will be rounded to nearest pixel.\n");
+    master_printf_stderr("Warning: grid volume is not an integer number of pixels; cell size will "
+                         "be rounded to nearest pixel.\n");
   return grid_volume(D1, a, 0, 0, (int)(zsize * a + 0.5));
 }
 
 grid_volume voltwo(double xsize, double ysize, double a) {
   if (!isinteger(xsize * a) || !isinteger(ysize * a))
-    master_printf_stderr(
-      "Warning: grid volume is not an integer number of pixels; cell size will be rounded to nearest pixel.\n");
+    master_printf_stderr("Warning: grid volume is not an integer number of pixels; cell size will "
+                         "be rounded to nearest pixel.\n");
   return grid_volume(D2, a, (xsize == 0) ? 1 : (int)(xsize * a + 0.5),
                      (ysize == 0) ? 1 : (int)(ysize * a + 0.5), 0);
 }
@@ -931,8 +922,8 @@ grid_volume vol2d(double xsize, double ysize, double a) { return voltwo(xsize, y
 
 grid_volume vol3d(double xsize, double ysize, double zsize, double a) {
   if (!isinteger(xsize * a) || !isinteger(ysize * a) || !isinteger(zsize * a))
-    master_printf_stderr(
-      "Warning: grid volume is not an integer number of pixels; cell size will be rounded to nearest pixel.\n");
+    master_printf_stderr("Warning: grid volume is not an integer number of pixels; cell size will "
+                         "be rounded to nearest pixel.\n");
   return grid_volume(D3, a, (xsize == 0) ? 1 : (int)(xsize * a + 0.5),
                      (ysize == 0) ? 1 : (int)(ysize * a + 0.5),
                      (zsize == 0) ? 1 : (int)(zsize * a + 0.5));
@@ -940,8 +931,8 @@ grid_volume vol3d(double xsize, double ysize, double zsize, double a) {
 
 grid_volume volcyl(double rsize, double zsize, double a) {
   if (!isinteger(rsize) || !isinteger(zsize))
-    master_printf_stderr(
-      "Warning: grid volume is not an integer number of pixels; cell size will be rounded to nearest pixel.\n");
+    master_printf_stderr("Warning: grid volume is not an integer number of pixels; cell size will "
+                         "be rounded to nearest pixel.\n");
 
   if (zsize == 0.0)
     return grid_volume(Dcyl, a, (int)(rsize * a + 0.5), 0, 1);
@@ -949,84 +940,6 @@ grid_volume volcyl(double rsize, double zsize, double a) {
     return grid_volume(Dcyl, a, (int)(rsize * a + 0.5), 0, (int)(zsize * a + 0.5));
 }
 
-grid_volume grid_volume::split(size_t n, int which) const {
-  if (n > nowned_min()) abort("Cannot split %zd grid points into %zd parts\n", nowned_min(), n);
-  if (n == 1) return *this;
-
-  // Try to get as close as we can...
-  int biglen = 0;
-  for (int i = 0; i < 3; i++)
-    if (num[i] > biglen) biglen = num[i];
-  const int split_point = (int)(biglen * (n / 2) / (double)n + 0.5);
-  const int num_low = (int)(split_point * n / (double)biglen + 0.5);
-  if (which < num_low)
-    return split_at_fraction(false, split_point).split(num_low, which);
-  else
-    return split_at_fraction(true, split_point).split(n - num_low, which - num_low);
-}
-
-grid_volume grid_volume::split_by_effort(int n, int which, int Ngv, const grid_volume *v,
-                                         double *effort) const {
-  const size_t grid_points_owned = nowned_min();
-  if (size_t(n) > grid_points_owned)
-    abort("Cannot split %zd grid points into %d parts\n", nowned_min(), n);
-  if (n == 1) return *this;
-  int biglen = 0;
-  direction splitdir = NO_DIRECTION;
-  LOOP_OVER_DIRECTIONS(dim, d) {
-    if (num_direction(d) > biglen) {
-      biglen = num_direction(d);
-      splitdir = d;
-    }
-  }
-  double best_split_measure = 1e20, left_effort_fraction = 0;
-  int best_split_point = 0;
-  vec corner = zero_vec(dim);
-  LOOP_OVER_DIRECTIONS(dim, d) {
-    corner.set_direction(d, origin.in_direction(d) + num_direction(d) / a);
-  }
-
-  for (int split_point = 1; split_point < biglen; split_point += 1) {
-    grid_volume v_left = *this;
-    v_left.set_num_direction(splitdir, split_point);
-    grid_volume v_right = *this;
-    v_right.set_num_direction(splitdir, num_direction(splitdir) - split_point);
-    v_right.shift_origin(splitdir, split_point * 2);
-
-    double total_left_effort = 0, total_right_effort = 0;
-    grid_volume vol;
-    if (Ngv == 0) {
-      total_left_effort = v_left.ntot();
-      total_right_effort = v_right.ntot();
-    }
-    else {
-      for (int j = 0; j < Ngv; j++) {
-        if (v_left.intersect_with(v[j], &vol)) total_left_effort += effort[j] * vol.ntot();
-        if (v_right.intersect_with(v[j], &vol)) total_right_effort += effort[j] * vol.ntot();
-      }
-    }
-    double split_measure = max(total_left_effort / (n / 2), total_right_effort / (n - n / 2));
-    if (split_measure < best_split_measure) {
-      best_split_measure = split_measure;
-      best_split_point = split_point;
-      left_effort_fraction = total_left_effort / (total_left_effort + total_right_effort);
-    }
-  }
-  const int split_point = best_split_point;
-
-  const int num_low = (size_t)(left_effort_fraction * n + 0.5);
-  // Revert to split() when effort method gives less grid points than chunks
-  if (size_t(num_low) > best_split_point * (grid_points_owned / biglen) ||
-      size_t(n - num_low) > (grid_points_owned - best_split_point * (grid_points_owned / biglen)))
-    return split(n, which);
-
-  if (which < num_low)
-    return split_at_fraction(false, split_point).split_by_effort(num_low, which, Ngv, v, effort);
-  else
-    return split_at_fraction(true, split_point)
-        .split_by_effort(n - num_low, which - num_low, Ngv, v, effort);
-}
-
 double grid_volume::get_cost() const {
   geom_box box = meep_geom::gv2box(surroundings());
   meep_geom::fragment_stats fstats(box);
@@ -1036,32 +949,54 @@ double grid_volume::get_cost() const {
 
 // return complex(left cost, right cost).  Should really be a tuple, but we don't want to require
 // C++11? yet?
-std::complex<double> grid_volume::get_split_costs(direction d, int split_point) const {
+std::complex<double> grid_volume::get_split_costs(direction d, int split_point,
+                                                  bool fragment_cost) const {
   double left_cost = 0, right_cost = 0;
   if (split_point > 0) {
     grid_volume v_left = *this;
     v_left.set_num_direction(d, split_point);
-    left_cost = v_left.get_cost();
+    left_cost = fragment_cost ? v_left.get_cost() : v_left.nowned_min();
   }
   if (split_point < num_direction(d)) {
     grid_volume v_right = *this;
     v_right.set_num_direction(d, num_direction(d) - split_point);
     v_right.shift_origin(d, split_point * 2);
-    right_cost = v_right.get_cost();
+    right_cost = fragment_cost ? v_right.get_cost() : v_right.nowned_min();
   }
   return std::complex<double>(left_cost, right_cost);
 }
 
 static double cost_diff(int desired_chunks, std::complex<double> costs) {
-  double left_cost = real(costs), right_cost = imag(costs);
-  return right_cost - left_cost * (desired_chunks - 1);
+  double left_cost = real(costs) / (desired_chunks / 2);
+  double right_cost = imag(costs) / (desired_chunks - desired_chunks / 2);
+  return right_cost - left_cost;
 }
 
-void grid_volume::find_best_split(int desired_chunks, int &best_split_point,
+void grid_volume::tile_split(int &best_split_point, direction &best_split_direction) const {
+  const size_t ntot_thresh = 10;
+  if (ntot() < ntot_thresh) {
+    best_split_point = 0;
+    best_split_direction = NO_DIRECTION;
+  }
+  else if (nx() > 1) {
+    best_split_point = nx() / 2;
+    best_split_direction = X;
+  }
+  else if (ny() > 1) {
+    best_split_point = ny() / 2;
+    best_split_direction = Y;
+  }
+  else {
+    best_split_point = nz() / 2;
+    best_split_direction = Z;
+  }
+}
+
+void grid_volume::find_best_split(int desired_chunks, bool fragment_cost, int &best_split_point,
                                   direction &best_split_direction,
                                   double &left_effort_fraction) const {
   if (size_t(desired_chunks) > nowned_min()) {
-    abort("Cannot split %zd grid points into %d parts\n", nowned_min(), desired_chunks);
+    meep::abort("Cannot split %zd grid points into %d parts\n", nowned_min(), desired_chunks);
   }
 
   left_effort_fraction = 0;
@@ -1083,7 +1018,7 @@ void grid_volume::find_best_split(int desired_chunks, int &best_split_point,
     int first = 0, last = num_direction(d);
     while (first < last) { // bisection search for balanced splitting
       int mid = (first + last) / 2;
-      double mid_diff = cost_diff(desired_chunks, get_split_costs(d, mid));
+      double mid_diff = cost_diff(desired_chunks, get_split_costs(d, mid, fragment_cost));
       if (mid_diff > 0) {
         if (first == mid) break;
         first = mid;
@@ -1094,78 +1029,39 @@ void grid_volume::find_best_split(int desired_chunks, int &best_split_point,
         break;
     }
     int split_point = (first + last) / 2;
-    std::complex<double> costs = get_split_costs(d, split_point);
+    std::complex<double> costs = get_split_costs(d, split_point, fragment_cost);
     double left_cost = real(costs), right_cost = imag(costs);
     double total_cost = left_cost + right_cost;
-    double split_measure = max(left_cost * (desired_chunks - 1), right_cost);
+    double split_measure = std::max(left_cost / (desired_chunks / 2),
+                                    right_cost / (desired_chunks - (desired_chunks / 2)));
+    // Give a 30% preference to the longest axis, as a heuristic to prefer lower communication costs
+    // when the split_measure is somewhat close.   TODO: use a data-driven communication cost
+    // function.
+    if (d == longest_axis) split_measure *= 0.7;
     if (split_measure < best_split_measure) {
-      if (d == longest_axis || split_measure < (best_split_measure - (0.3 * best_split_measure))) {
-        // Only use this split_measure if we're on the longest_axis, or if the split_measure is
-        // more than 30% better than the best_split_measure. This is a heuristic to prefer lower
-        // communication costs when the split_measure is somewhat close.
-        // TODO: Use machine learning to get a cost function for the communication instead of hard
-        // coding 0.3
-
-        best_split_measure = split_measure;
-        best_split_point = split_point;
-        best_split_direction = d;
-        left_effort_fraction = left_cost / total_cost;
-      }
+      best_split_measure = split_measure;
+      best_split_point = split_point;
+      best_split_direction = d;
+      left_effort_fraction = left_cost / total_cost;
     }
   }
 }
 
-std::vector<grid_volume> grid_volume::split_into_n(int n) const {
-  std::vector<grid_volume> result;
-  grid_volume rest_gv = *this;
-  while (n > 1) {
-    int best_split_point;
-    direction best_split_direction;
-    double effort_fraction;
-    rest_gv.find_best_split(n, best_split_point, best_split_direction, effort_fraction);
-    int num_in_split_dir = num_direction(best_split_direction);
-    result.push_back(
-        rest_gv.split_at_fraction(false, best_split_point, best_split_direction, num_in_split_dir));
-    rest_gv =
-        rest_gv.split_at_fraction(true, best_split_point, best_split_direction, num_in_split_dir);
-    --n;
-  }
-  result.push_back(rest_gv);
-  return result;
-}
-
-grid_volume grid_volume::split_at_fraction(bool want_high, int numer, int bestd,
-                                           int bestlen) const {
-
-  if (bestd == -1) {
-    for (int i = 0; i < 3; i++)
-      if (num[i] > bestlen) {
-        bestd = i;
-        bestlen = num[i];
-      }
-  }
-  else {
-    bestd %= 3;
-  }
-
-  if (bestd == -1) {
-    for (int i = 0; i < 3; i++)
-      master_printf("num[%d] = %d\n", i, num[i]);
-    abort("Crazy weird splitting error.\n");
-  }
+grid_volume grid_volume::split_at_fraction(bool side_high, int split_pt, int split_dir) const {
+  if (dim == Dcyl) split_dir %= 3;
   grid_volume retval(dim, a, 1, 1, 1);
   for (int i = 0; i < 3; i++)
     retval.num[i] = num[i];
-  if (numer >= num[bestd]) abort("Aaack bad bug in split_at_fraction.\n");
-  direction d = (direction)bestd;
+  if (split_pt >= num[split_dir]) meep::abort("Aaack bad bug in split_at_fraction.\n");
+  direction d = (direction)split_dir;
   if (dim == Dcyl && d == X) d = R;
   retval.set_origin(io);
-  if (want_high) retval.shift_origin(d, numer * 2);
+  if (side_high) retval.shift_origin(d, split_pt * 2);
 
-  if (want_high)
-    retval.num[bestd] -= numer;
+  if (side_high)
+    retval.num[split_dir] -= split_pt;
   else
-    retval.num[bestd] = numer;
+    retval.num[split_dir] = split_pt;
   retval.num_changed();
   return retval;
 }
@@ -1173,8 +1069,8 @@ grid_volume grid_volume::split_at_fraction(bool want_high, int numer, int bestd,
 // Halve the grid_volume for symmetry exploitation...must contain icenter!
 grid_volume grid_volume::halve(direction d) const {
   grid_volume retval(*this);
-  // note that icenter-io is always even by construction of grid_volume::icenter
-  retval.set_num_direction(d, (icenter().in_direction(d) - io.in_direction(d)) / 2);
+  retval.set_num_direction(d, 1 + (big_corner().in_direction(d) - icenter().in_direction(d)) / 2);
+  retval.set_origin(d, icenter().in_direction(d) - 2);
   return retval;
 }
 
@@ -1195,12 +1091,12 @@ ivec grid_volume::icenter() const {
      symmetry point, and therefore icenter-io must be *even*
      in all components in order that rotations preserve the Yee lattice. */
   switch (dim) {
-    case D1: return io + ivec(nz()).round_up_to_even();
-    case D2: return io + ivec(nx(), ny()).round_up_to_even();
-    case D3: return io + ivec(nx(), ny(), nz()).round_up_to_even();
-    case Dcyl: return io + iveccyl(0, nz()).round_up_to_even();
+    case D1: return io + ivec(nz()).round_down_to_even();
+    case D2: return io + ivec(nx(), ny()).round_down_to_even();
+    case D3: return io + ivec(nx(), ny(), nz()).round_down_to_even();
+    case Dcyl: return io + iveccyl(0, nz()).round_down_to_even();
   }
-  abort("Can't do symmetry with these dimensions.\n");
+  meep::abort("Can't do symmetry with these dimensions.\n");
   return ivec(0); // This is never reached.
 }
 
@@ -1208,7 +1104,7 @@ vec grid_volume::center() const { return operator[](icenter()); }
 
 symmetry rotate4(direction axis, const grid_volume &gv) {
   symmetry s = identity();
-  if (axis > 2) abort("Can only rotate4 in 2D or 3D.\n");
+  if (axis > 2) meep::abort("Can only rotate4 in 2D or 3D.\n");
   s.g = 4;
   FOR_DIRECTIONS(d) {
     s.S[d].d = d;
@@ -1224,7 +1120,7 @@ symmetry rotate4(direction axis, const grid_volume &gv) {
 
 symmetry rotate2(direction axis, const grid_volume &gv) {
   symmetry s = identity();
-  if (axis > 2) abort("Can only rotate2 in 2D or 3D.\n");
+  if (axis > 2) meep::abort("Can only rotate2 in 2D or 3D.\n");
   s.g = 2;
   s.S[(axis + 1) % 3].flipped = true;
   s.S[(axis + 2) % 3].flipped = true;
@@ -1316,6 +1212,14 @@ int symmetry::multiplicity() const {
     return g;
 }
 
+int symmetry::multiplicity(ivec &x) const {
+  int m = multiplicity();
+  for (int n = 1; n < m; ++n) {
+    if (transform(x, n) == x) return n;
+  }
+  return m;
+}
+
 symmetry symmetry::operator+(const symmetry &b) const {
   // The following optimization ignores identity when adding symmetries
   // together.  This is important because identity has an undefined
@@ -1376,9 +1280,7 @@ signed_direction symmetry::transform(direction d, int n) const {
       else
         return next->transform(sd.d, nrest) * ph;
     }
-    else {
-      return sd * ph;
-    }
+    else { return sd * ph; }
   }
 }
 
@@ -1571,18 +1473,19 @@ volume_list *symmetry::reduce(const volume_list *gl) const {
 
 /***************************************************************************/
 
-static double poynting_fun(const complex<double> *fields, const vec &loc, void *data_) {
+static double poynting_fun(const complex<realnum> *fields, const vec &loc, void *data_) {
   (void)loc;   // unused
   (void)data_; // unused
-  return (real(conj(fields[0]) * fields[1]) - real(conj(fields[2]) * fields[3]));
+  return (real(conj(cdouble(fields[0])) * cdouble(fields[1])) -
+          real(conj(cdouble(fields[2])) * cdouble(fields[3])));
 }
 
-static double energy_fun(const complex<double> *fields, const vec &loc, void *data_) {
+static double energy_fun(const complex<realnum> *fields, const vec &loc, void *data_) {
   (void)loc; // unused
   int nfields = *((int *)data_) / 2;
   double sum = 0;
   for (int k = 0; k < nfields; ++k)
-    sum += real(conj(fields[2 * k]) * fields[2 * k + 1]);
+    sum += real(conj(cdouble(fields[2 * k])) * cdouble(fields[2 * k + 1]));
   return sum * 0.5;
 }
 
@@ -1638,10 +1541,10 @@ field_rfunction derived_component_func(derived_component c, const grid_volume &g
             cs[nfields++] = direction_component(Bx, component_direction(c0));
           }
         }
-      if (nfields > 12) abort("too many field components");
+      if (nfields > 12) meep::abort("too many field components");
       return energy_fun;
 
-    default: abort("unknown derived_component in derived_component_func");
+    default: meep::abort("unknown derived_component in derived_component_func");
   }
   return 0;
 }
@@ -1690,8 +1593,7 @@ const char *volume::str(char *buffer, size_t buflen) {
     buffer = sbuf;
     buflen = sizeof(sbuf);
   }
-  snprintf(buffer, buflen, "min_corner:%s, max_corner:%s",
-           min_corner.str(), max_corner.str());
+  snprintf(buffer, buflen, "min_corner:%s, max_corner:%s", min_corner.str(), max_corner.str());
   return buffer;
 }
 
@@ -1703,45 +1605,45 @@ const char *grid_volume::str(char *buffer, size_t buflen) {
     buflen = sizeof(sbuf);
   }
 
-  written += snprintf(buffer+written, buflen-written,
+  written += snprintf(buffer + written, buflen - written,
                       "grid_volume {\n  dim:%s, a:%f, inva:%f, num:{%d, %d, %d}\n",
                       dimension_name(dim), a, inva, num[0], num[1], num[2]);
 
   // Adapted from the print() method
   LOOP_OVER_DIRECTIONS(dim, d) {
-    written += snprintf(buffer+written, buflen-written,
-              "  %s =%5g - %5g (%5g) \t", direction_name(d), origin.in_direction(d),
-              origin.in_direction(d) + num_direction(d) / a, num_direction(d) / a);
-    if (buflen-written <=0)
-        break;
+    written += snprintf(buffer + written, buflen - written, "  %s =%5g - %5g (%5g) \t",
+                        direction_name(d), origin.in_direction(d),
+                        origin.in_direction(d) + num_direction(d) / a, num_direction(d) / a);
+    if (buflen - written <= 0) break;
   }
-  snprintf(buffer+written, buflen-written, "\n}");
+  snprintf(buffer + written, buflen - written, "\n}");
   return buffer;
 }
 
-
 /********************************************************************/
 /********************************************************************/
 /********************************************************************/
-grid_volume grid_volume::subvolume(ivec is, ivec ie) {
-  if (!(contains(is) && contains(ie))) abort("invalid extents in subvolume");
+grid_volume grid_volume::subvolume(ivec is, ivec ie, component c) {
+  if (!(contains(is) && contains(ie))) meep::abort("invalid extents in subvolume");
   grid_volume sub;
   sub.dim = dim;
   sub.a = a;
   sub.inva = inva;
-  sub.init_subvolume(is, ie);
+  sub.init_subvolume(is, ie, c);
   return sub;
 }
 
-void grid_volume::init_subvolume(ivec is, ivec ie) {
+void grid_volume::init_subvolume(ivec is, ivec ie, component c) {
   ivec origin(dim, 0);
+  for (int i = 0; i < 3; i++)
+    num[i] = 0;
   LOOP_OVER_DIRECTIONS(dim, d) {
     num[(int)d] = (ie - is).in_direction(d) / 2;
-    origin.set_direction(d, is.in_direction(d));
+    origin.set_direction(d, is.in_direction(d) - iyee_shift(c).in_direction(d));
   }
   num_changed();
-  center_origin();
-  shift_origin(origin);
+  // center_origin();
+  set_origin(origin);
 }
 
 } // namespace meep
